# Phantom script by NowakPL <nowak@xpam.de>

local ScenarioUtils = import('/lua/sim/ScenarioUtilities.lua')
local PhantomConfig = {}
local IsPhantom = {}
local PhantomIsOn = false

function OnPopulate()
	ScenarioUtils.InitializeArmies()
end

function ListPlayerArmies()
	local player_armies = {}
	for army, name in ListArmies() do
		if ArmyIsCivilian(army) == false and ArmyIsOutOfGame(army) == false then
			table.insert(player_armies, army)
		end
	end
	return player_armies
end

function RandomPlayerArmy()
	local player_armies = ListPlayerArmies()
	return player_armies[Random(table.getn(player_armies))]
end

function SelectPhantoms()
	local player_armies = ListPlayerArmies()
	for i = 1, PhantomConfig.phantoms do
		if table.getn(player_armies) >= 1 then
			local army = table.remove(player_armies, Random(table.getn(player_armies)))
			IsPhantom[army] = true
			local brain = GetArmyBrain(army)
			brain:GiveStorage('MASS', 2000)
			brain:GiveStorage('ENERGY', 10000)
		end
	end
end

function PhantomNotifyThread()
	if IsPhantom[GetFocusArmy()] == true then
		PrintText('You are the Phantom!', 20, 'ffff2020', 15, 'center')
	else
		PrintText('Identify and kill the Phantom!', 20, 'ffffff00', 15, 'center')
	end
	WaitSeconds(0.1)
	while true do
		local phantoms = ''
		local phantoms_count = 0
		for i, army in ListPlayerArmies() do
			if IsPhantom[army] == true then
				local nickname = GetArmyBrain(army).Nickname
				if phantoms == '' then
					phantoms = nickname
				else
					phantoms = phantoms .. ', ' .. nickname
				end
				phantoms_count = phantoms_count + 1
			end
		end
		if IsPhantom[GetFocusArmy()] == true then
			if phantoms_count == 1 then
				PrintText('Phantom: ' .. phantoms, 16, nil, 13, 'lefttop')
			elseif phantoms_count >= 2 then
				PrintText('Phantoms: ' .. phantoms, 16, nil, 13, 'lefttop')
			end
		else
			if phantoms_count == 1 then
				PrintText('1 Phantom left to kill', 16, nil, 13, 'lefttop')
			elseif phantoms_count >= 2 then
				PrintText(phantoms_count .. ' Phantoms left to kill', 16, nil, 13, 'lefttop')
			end
		end
		WaitSeconds(15)
	end
end

function PhantomStartThread()
	WaitSeconds(PhantomConfig.start_time)
	SelectPhantoms()
	local phantom_notify_thread = ForkThread(PhantomNotifyThread)
	PhantomIsOn = true
end

function PhantomMainThread()
	local old_phantoms_alive = -1
	local old_players_alive = -1
	local on_start_players_alive = -1
	local end_game = false
	while end_game == false do
		local players_mass = 0
		local players_energy = 0
		local players_alive = 0
		local phantoms_alive = 0
		for i, army in ListPlayerArmies() do
			SetAlliedVictory(army, false)
			local brain = GetArmyBrain(army)
			if IsPhantom[army] == true then
				phantoms_alive = phantoms_alive + 1
			else
				players_mass = players_mass + brain:GetEconomyIncome('MASS')
				players_energy = players_energy + brain:GetEconomyIncome('ENERGY')
				players_alive = players_alive + 1
			end
		end
		if PhantomIsOn != true then
			if players_alive <= PhantomConfig.phantoms then
				PrintText('Not enough players to continue :(', 20, nil, 10, 'center')
				end_game = true
			end
		else
			if on_start_players_alive == -1 then
				on_start_players_alive = players_alive
			end
			if old_players_alive >= 0 and old_players_alive != players_alive and players_alive >= 1 then
				if (PhantomConfig.phantom_mass_bonus + PhantomConfig.phantom_energy_bonus) != 0.0 then
					PrintText('Innocent player has been killed! The Phantom(s) receive resource boost!', 20, nil, 10, 'center')
				else
					PrintText('Innocent player has been killed!', 20, nil, 10, 'center')
				end
			end
			if old_phantoms_alive >= 0 and old_phantoms_alive != phantoms_alive and phantoms_alive >= 1 then
				PrintText('One of the Phantoms has been killed!', 20, nil, 10, 'center')
			end
			if phantoms_alive >= 1 and players_alive == 0 then
				PrintText('The Phantom(s) win!', 20, nil, 10, 'center')
				end_game = true
			end
			if phantoms_alive == 0 and players_alive >= 1 then
				PrintText('The Phantom(s) lose!', 20, nil, 10, 'center')
				end_game = true
			end
			if phantoms_alive == 0 and players_alive == 0 then
				PrintText('The game ends in a draw!', 20, nil, 10, 'center')
				end_game = true
			end
			if end_game == false then
				local bonus = on_start_players_alive - players_alive
				local mult = 1.0 / phantoms_alive
				for i, army in ListPlayerArmies() do
					if IsPhantom[army] == true then
						local brain = GetArmyBrain(army)
						brain:GiveResource('MASS', players_mass * mult * (PhantomConfig.phantom_mass_mult + bonus * PhantomConfig.phantom_mass_bonus))
						brain:GiveResource('ENERGY', players_energy * mult * (PhantomConfig.phantom_energy_mult + bonus * PhantomConfig.phantom_energy_bonus))
					end
				end
			end
		end
		old_phantoms_alive = phantoms_alive
		old_players_alive = players_alive
		WaitSeconds(0.1)
	end
	for army, name in ListArmies() do
		local brain = GetArmyBrain(army)
		if not brain:IsDefeated() then
			brain:OnVictory()
		end
	end
	WaitSeconds(2)
	EndGame()
end

function OnStart(self)
	PhantomConfig.start_time = 8 * 60
	
	if ScenarioInfo.Options.Timeouts == '0' then # none
		PhantomConfig.phantoms = 1
	elseif ScenarioInfo.Options.Timeouts == '3' then # three
		PhantomConfig.phantoms = 2
	else # infinite
		PhantomConfig.phantoms = 3
	end
	if ScenarioInfo.Options.CivilianAlliance == 'removed' then # none
		PhantomConfig.phantom_mass_mult = 0.3
		PhantomConfig.phantom_energy_mult = 0.3
		PhantomConfig.phantom_mass_bonus = 0.0
		PhantomConfig.phantom_energy_bonus = 0.0
	elseif ScenarioInfo.Options.CivilianAlliance == 'neutral' then # neutral
		PhantomConfig.phantom_mass_mult = 0.3
		PhantomConfig.phantom_energy_mult = 0.3
		PhantomConfig.phantom_mass_bonus = 0.1
		PhantomConfig.phantom_energy_bonus = 0.1
	else # enemy
		PhantomConfig.phantom_mass_mult = 0.5
		PhantomConfig.phantom_energy_mult = 0.5
		PhantomConfig.phantom_mass_bonus = 0.0
		PhantomConfig.phantom_energy_bonus = 0.0
	end
	
	if ScenarioInfo.Options.TeamLock == 'unlocked' then
		for army, name in ListArmies() do
			IsPhantom[army] = false
			local brain = GetArmyBrain(army)
			brain.CalculateScore = function(this_brain)
				return math.floor(this_brain:GetArmyStat("UnitCap_Current", 0.0).Value)
			end
		end

		local player_armies = ListPlayerArmies()
		for i, army in player_armies do
			SetAlliedVictory(army, false)
			for i2, army2 in player_armies do
				SetAlliance(army, army2, 'Ally')
			end
		end

		local phantom_main_thread = ForkThread(PhantomMainThread)
		local phantom_start_thread = ForkThread(PhantomStartThread)
	end
end
